/*
 * xmms-mad - mp3 plugin for xmms
 * Copyright (C) 2001-2002 Sam Clegg - See COPYING
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: xmms-mad.c,v 1.15 2005/02/10 14:06:23 sam Exp $
 */
#include "config.h"
#include "xmms-mad.h"
#include "input.h"

#include <gtk/gtk.h>
#include <audacious/util.h>
#include <audacious/configdb.h>
#include <stdarg.h>
#include <fcntl.h>

/*
 * Global variables
 */
struct xmmsmad_config_t xmmsmad_config;   /**< global configuration */
static GStaticMutex mutex;

/*
 * static variables
 */
static GThread *decode_thread; /**< the single decoder thread */
static struct mad_info_t info;   /**< info for current track */

#ifndef NOGUI
static GtkWidget *about_window = 0;
static GtkWidget *error_dialog = 0;
#endif /* !NOGUI */

static void
xmmsmad_init ()
{
	ConfigDb *db;
  xmmsmad_config.http_buffer_size = HTTP_BUFFER_SIZE,
  xmmsmad_config.fast_play_time_calc = TRUE;
  xmmsmad_config.use_xing = TRUE;
  xmmsmad_config.dither = TRUE;
  db = bmp_cfg_db_open();
  if(db)
    {
      bmp_cfg_db_get_int(db, "MAD", "http_buffer_size", &xmmsmad_config.http_buffer_size);
      bmp_cfg_db_get_bool(db, "MAD", "fast_play_time_calc", &xmmsmad_config.fast_play_time_calc);
      bmp_cfg_db_get_bool(db, "MAD", "use_xing", &xmmsmad_config.use_xing);
      bmp_cfg_db_get_bool(db, "MAD", "dither", &xmmsmad_config.dither);
    }
  bmp_cfg_db_close(db);
  g_static_mutex_init(&mutex);
}

static void
xmmsmad_cleanup ()
{
}

static void
xmmsmad_set_eq (int on, float preamp, float *bands)
{
  //printf("set_eq\n");
}

static int
xmmsmad_is_our_file (char *filename)
{
  int fin = 0;
  int rtn = 0;
  guchar check [4];

  if (strncasecmp ("http://", filename, 7) == 0)
    {
      if (strcasecmp ("mp3", filename + strlen (filename) - 3) == 0)
        {
          /* [FIXME] do some checks here to make sure we have an mp3 stream */
          rtn = 1;
        }
    }
  else
    {
      fin = open (filename, O_RDONLY);
      if (fin >= 0 && read (fin, check, 4) == 4)
	{
	  /* If first two bytes are a sync header or three bytes are "ID3" */
	  if ( (check[0] == 0xff && (check[1] & 0x70) == 0x70) 
	       || memcmp (check, "ID3", 3) == 0) 
	    {
	      rtn = 1;
	    }
	  else if (memcmp (check, "RIFF", 4) == 0) /* RIFF data */
	    {        
	      lseek (fin, 4, SEEK_CUR);
	      read (fin, check, 4);

	      if (memcmp (check, "RMP3", 4) == 0)
		{
		  rtn = 1;
		}
#if 0
	      /* FIXME: make this test endian safe */
	      else if (memcmp (check, "WAVE"))
	        {
		  /* seek to codec ID and see if it's MP3 */
		  lseek (fin, 20, SEEK_SET);
		  check = 0L;
		  read (fin, &check, 2);
		  if (check == 0x55)
		    {
		      rtn = 1;
		    }
		}
#endif 
	    }
	}
    }

  close (fin);
  return rtn;
}

static void
xmmsmad_stop ()
{
#ifdef DEBUG
  g_message ("f: xmmsmad_stop");
#endif /* DEBUG */
  g_static_mutex_lock (&mutex);
  if (decode_thread)
    {
	    info.stop=1;
#ifdef DEBUG
      g_message ("waiting for thread");
#endif /* DEBUG */
      g_thread_join(decode_thread);
#ifdef DEBUG
      g_message ("thread done");
#endif /* DEBUG */
      input_term (&info);
      decode_thread = NULL;
      info.stop = 0;
    }
  g_static_mutex_unlock (&mutex);
}


static void
xmmsmad_play_file (char *url)
{
  gboolean rtn;

  xmmsmad_stop ();

#ifdef DEBUG
  g_message ("playing %s", url);
#endif /* DEBUG */

  if (input_init (&info, url) == FALSE)
    {
      g_message ("error initialising input");
      return;
    }

  rtn = input_get_info (&info, xmmsmad_config.fast_play_time_calc);

  if (rtn == FALSE)
    {
      g_message ("error reading input info");
      return;
    }
  decode_thread = g_thread_create(decode, (void *) &info, TRUE, NULL);
}

static void
xmmsmad_pause (short paused)
{
  mad_plugin.output->pause (paused);
}

static void
xmmsmad_seek (int time)
{
  /* xmms gives us the desired seek time in seconds */
  info.seek = time;
}

static int
xmmsmad_get_time ()
{
  if (info.stop)
    {
#ifdef DEBUG
      g_message ("returning -1 from get_time");
#endif
      return -1;
    }
  return mad_plugin.output->output_time ();
}

/**
 * Scan the given file or URL.
 * Fills in the title string and the track length in milliseconds.
 */
static void
xmmsmad_get_song_info (char *url, char **title, int *length)
{
  struct mad_info_t myinfo;
#ifdef DEBUG
  g_message ("f: xmmsmad_get_song_info: %s", url);
#endif /* DEBUG */
  input_init (&myinfo, url);
  if (input_get_info (&myinfo, TRUE) == TRUE)
    {
      *title  = strdup (myinfo.title);
      *length = mad_timer_count (myinfo.duration, MAD_UNITS_MILLISECONDS);
    }
  else
    {
      *title = strdup (url);
      *length = -1;
    }
  input_term (&myinfo);
#ifdef DEBUG
  g_message ("e: xmmsmad_get_song_info");
#endif /* DEBUG */
}

static void
xmmsmad_about ()
{
    static GtkWidget *aboutbox;

    if (aboutbox != NULL)
        return;

    aboutbox = xmms_show_message("About MAD MPEG Decoder Plugin",
                                 "Based on libmad",
                                 "Ok",
                                 FALSE, NULL, NULL);

    g_signal_connect(G_OBJECT(aboutbox), "destroy",
                     G_CALLBACK(gtk_widget_destroyed), &aboutbox);
}



/**
 * Display a GTK box containing the given error message.
 * Taken from mpg123 plugin.
 */
void
xmmsmad_error (char *error, ...)
{
#ifndef NOGUI
  if (!error_dialog)
    {
      va_list args;
      char string[256];
      va_start (args, error);
      vsnprintf (string, 256, error, args);
      va_end (args);
      GDK_THREADS_ENTER ();
      error_dialog = xmms_show_message ("Error", string, "Ok", FALSE, 0, 0);
      gtk_signal_connect (GTK_OBJECT (error_dialog),
			  "destroy",
			  GTK_SIGNAL_FUNC (gtk_widget_destroyed),
			  &error_dialog);
      GDK_THREADS_LEAVE ();
    }
#endif /* !NOGUI */
}

extern void xmmsmad_get_file_info (char *filename);
extern void xmmsmad_configure ();

/**
 * XMMS input plugin hook.
 */

InputPlugin mad_plugin = {
	NULL,
	NULL,
	NULL,
	xmmsmad_init,
	xmmsmad_about,
	xmmsmad_configure,
	xmmsmad_is_our_file,
	NULL,
	xmmsmad_play_file,
	xmmsmad_stop,
	xmmsmad_pause,
	xmmsmad_seek,
	xmmsmad_set_eq,
	xmmsmad_get_time,
	NULL,
	NULL,
	xmmsmad_cleanup,
	NULL,
	NULL,
	NULL,
	NULL,
	xmmsmad_get_song_info,
	xmmsmad_get_file_info,
	NULL
};

#if 0
struct _InputPlugin {
    gpointer handle;
    gchar *filename;
    gchar *description;
    void (*init) (void);
    void (*about) (void);
    void (*configure) (void);
    gint (*is_our_file) (gchar * filename);
    GList *(*scan_dir) (gchar * dirname);
    void (*play_file) (gchar * filename);
    void (*stop) (void);
    void (*pause) (gshort paused);
    void (*seek) (gint time);
    void (*set_eq) (gint on, gfloat preamp, gfloat * bands);
    gint (*get_time) (void);
    void (*get_volume) (gint * l, gint * r);
    void (*set_volume) (gint l, gint r);
    void (*cleanup) (void);
    InputVisType (*get_vis_type) (void);
    void (*add_vis_pcm) (gint time, AFormat fmt, gint nch, gint length, gpointer ptr);
    void (*set_info) (gchar * title, gint length, gint rate, gint freq, gint nch);
    void (*set_info_text) (gchar * text);
    void (*get_song_info) (gchar * filename, gchar ** title, gint * length);
    void (*file_info_box) (gchar * filename);
    OutputPlugin *output;
};
#endif

InputPlugin *
get_iplugin_info(void)
{
	mad_plugin.description = DESCRIPTION " " VERSION;
	return &mad_plugin;
}
